import { getModelForClass, index, modelOptions, prop } from '@typegoose/typegoose'
import { dbconn } from 'decorators/dbconn'
import { BaseModule } from './Base'
import logger from 'logger/logger'
import { GeneralEvent } from './GeneralEvent'

const STAKE_EVENT = 'Staked'
const REDEEM_EVENT = 'Redeem'

@dbconn()
@index({ chain: 1, nft: 1, tokenId: 1, start: 1 }, { unique: true })
@index({ chain: 1, user: 1, nft: 1 }, { unique: false })
@modelOptions({
  schemaOptions: { collection: 'nft_stake_info', timestamps: true },
})
export class NftStakeClass extends BaseModule {
  @prop({ required: true })
  public chain: string
  @prop()
  public blockNumber: number
  // stake info
  @prop()
  public address: string
  @prop()
  public user: string
  @prop()
  public nft: string
  @prop()
  public tokenId: string
  @prop()
  public start: number
  @prop()
  public stakeTime: number
  // 1: staked, 2: redeemed
  @prop()
  public status: number

  @prop()
  public redeemTime: number

  @prop({ default: 0 })
  public version: number

  /**
   * 
    {
      user: '0x50A8e60041A206AcaA5F844a1104896224be6F39',
      infos: [
        {
          user: '0x50A8e60041A206AcaA5F844a1104896224be6F39',
          nft: '0x09F0dFFA584B1277D7c4E44265a6b5D03303Fc99',
          tokenId: '1',
          start: '1704870143',
          stakeTime: '2592000'
        }
      ]
    }
   */
  public static async parseEvent(event: typeof GeneralEvent) {
    let { user, infos } = event.decodedData
    const isStake = event.event === STAKE_EVENT
    const blockNumber = event.blockNumber
    const chain = event.chain
    if (infos.length === 0) {
      return
    }
    user = user?.toLowerCase()
    for (let info of infos) {
      let { nft, tokenId, start, stakeTime } = info
      nft = nft?.toLowerCase()
      let address = event.address?.toLowerCase()
      start = parseInt(start)
      stakeTime = parseInt(stakeTime)
      if (isStake) {
        logger.info(`stake nft: ${nft}, tokenId: ${tokenId}, user: ${user}, blockNumber: ${blockNumber}`)
        let record = new NftStake({
          address,
          chain,
          nft,
          tokenId,
          user,
          blockNumber,
          status: 1,
          start,
          stakeTime,
        })
        await record.save()
      } else {
        logger.info(`redeem nft: ${nft}, tokenId: ${tokenId}, user: ${user}, blockNumber: ${blockNumber}`)
        await NftStake.insertOrUpdate(
          { chain, nft, tokenId, start },
          {
            status: 2,
            address,
            user,
            redeemTime: (Date.now() / 1000) | 0,
            blockNumber,
            $inc: { version: 1 },
          },
        )
      }
    }
  }
}

export const NftStake = getModelForClass(NftStakeClass, {
  existingConnection: NftStakeClass['db'],
})
