import { IChain } from 'chain/allchain'
import { batchEthLogs, ethGetBlockByNumber, ethGetLogs, retryEthBlockNumber, retryEthGetBlockByNumber } from 'chain/chain.api'
import logger from 'logger/logger'
import { GeneralEvent } from 'models/GeneralEvent'

import { decodeEvent, getTopics } from 'zutils/utils/chain.util'

import { NftHolder } from 'models/NftHolder'
import { TokenHolder } from 'models/TokenHolder'
import { NftStake } from 'models/NftStake'
import { IEventCfg } from 'interface/IEventCfg'
import { ZRedisClient } from 'zutils'
import { formatDate } from 'utils/utcdate.util'
import { TreasureHunt } from 'models/TreasureHunt'

let eventProcessers = {
  NftHolder: NftHolder,
  TokenHolder: TokenHolder,
  NftStake: NftStake,
  TreasureHunt: TreasureHunt,
}

const INTERVAL = 100
export class EventBatchSvr {
  chainCfg: IChain
  eventCfgs: IEventCfg[] = []
  processer: Map<string, IEventCfg> = new Map()
  fromBlock: number = Number.MAX_SAFE_INTEGER
  redisKey = ''
  rpc = ''
  constructor(_chainCfg: IChain, _eventCfgs: IEventCfg[]) {
    this.chainCfg = _chainCfg
    this.eventCfgs = _eventCfgs
    this.rpc = _chainCfg.rpc.split('|')[0]
    for (let cfg of this.eventCfgs) {
      this.fromBlock = Math.min(this.fromBlock, cfg.fromBlock)
      if (!cfg.topic) {
        cfg.topic = getTopics(cfg.abi)
      }
      this.processer.set((cfg.address + cfg.topic).toLowerCase(), cfg)
    }
    this.eventCfgs.sort((a, b) => a.topic.localeCompare(b.topic))
    this.redisKey = `event_${this.chainCfg.id}`
  }

  async divQueryPassEvents(cfg: IEventCfg, fromBlock: number, toBlock: number) {
    const middle = ((fromBlock + toBlock) / 2) | 0
    const middlePlusOne = middle + 1
    const firstHalfEvents = await this.getPastEvents(cfg, fromBlock, middle)
    const secondHalfEvents = await this.getPastEvents(cfg, middlePlusOne, toBlock)
    return [...firstHalfEvents, ...secondHalfEvents]
  }

  async getPastEvents(cfg: IEventCfg, fromBlock: number, toBlock: number) {
    let events = []
    logger.info(`get past events: ${cfg.address}, ${cfg.event}, ${fromBlock}, ${toBlock}`)
    try {
      const _param = this.buildQueryParams(cfg, fromBlock, toBlock)
      const result = await ethGetLogs(this.rpc, _param.params)
      if (result.error) {
        if (result.error.message && /Log response size exceeded/.test(result.error.message)) {
          events = await this.divQueryPassEvents(cfg, fromBlock, toBlock)
        } else {
          throw result.error
        }
        logger.info('fetch history error: ', result.error.message || result.error)
      } else {
        events = result.result
      }
    } catch (e) {
      if (e.message && /Log response size exceeded/.test(e.message)) {
        events = await this.divQueryPassEvents(cfg, fromBlock, toBlock)
      }
    }
    return events
  }
  async execute() {
    let currentBlock = await retryEthBlockNumber(this.rpc)
    let toBlock = parseInt(currentBlock.result, 16)
    let blockStr = await new ZRedisClient().get(this.redisKey)
    if (blockStr) {
      this.fromBlock = Math.max(parseInt(blockStr), this.fromBlock)
    }
    logger.info(`sync events with chain: ${this.chainCfg.id}, from: ${this.fromBlock}, to: ${toBlock}`)
    let uninParams = []
    let topicsSet = new Set()
    for (let cfg of this.eventCfgs) {
      await this.fixBlockNumber(cfg, this.chainCfg, toBlock)
      if (cfg.fromBlock === this.fromBlock && cfg.toBlock === toBlock) {
        let _param = uninParams[uninParams.length - 1]
        if (!_param || !topicsSet.has(cfg.topic)) {
          _param = this.buildQueryParams(cfg)
          uninParams.push(_param)
          topicsSet = new Set()
          topicsSet.add(cfg.topic)
        } else {
          if (!topicsSet.has(cfg.topic)) {
            _param.params[0].topics.push(cfg.topic)
            topicsSet.add(cfg.topic)
          }
          _param.params[0].address.push(cfg.address)
        }
      } else {
        const historyEvents = await this.getPastEvents(cfg, cfg.fromBlock, toBlock)
        await this.processEvents(historyEvents)
      }
    }
    let nextBlock = this.fromBlock
    if (uninParams.length > 1) {
      logger.debug(`unin params: ${uninParams.length}`)
      let results = await batchEthLogs(this.rpc, uninParams)
      for (let i = 0; i < results.length; i++) {
        if (results[i].error) {
          throw results[i].error
        }
        let events = results[i].result
        await this.processEvents(events)
      }
    } else if (uninParams.length == 1) {
      logger.debug(`unin params length 1, use eth_getLogs`)
      let result = await ethGetLogs(this.rpc, uninParams[0].params)
      if (result.error) {
        throw result.error
      }
      await this.processEvents(result.result)
    }
    nextBlock = toBlock + 1
    for (let cfg of this.eventCfgs) {
      cfg.fromBlock = nextBlock
      cfg.toBlock = 0
      const redisKey = this.buildRedisKey(cfg)
      await new ZRedisClient().set(redisKey, cfg.fromBlock + '')
    }
    await new ZRedisClient().set(this.redisKey, nextBlock + '')
  }

  buildRedisKey(cfg: IEventCfg) {
    return `event_${this.chainCfg.id}_${cfg.address}_${cfg.event}`
  }

  async fixBlockNumber(cfg: IEventCfg, chainCfg: IChain, toBlock: number) {
    const redisKey = this.buildRedisKey(cfg)
    let blockStr = await new ZRedisClient().get(redisKey)
    if (blockStr) {
      cfg.fromBlock = Math.max(parseInt(blockStr), cfg.fromBlock)
    }
    if (chainCfg.maxEvents) {
      cfg.toBlock = Math.min(cfg.fromBlock + chainCfg.maxEvents, toBlock)
    } else {
      cfg.toBlock = toBlock
    }
  }

  buildQueryParams(cfg: IEventCfg, fromBlock?: number, toBlock?: number) {
    fromBlock = fromBlock || cfg.fromBlock
    toBlock = toBlock || cfg.toBlock
    const params: any = {
      fromBlock: '0x' + fromBlock.toString(16),
      toBlock: toBlock ? '0x' + toBlock?.toString(16) : 'latest',
      address: [cfg.address],
    }
    params.topics = [cfg.topic]
    const result = {
      jsonrpc: '2.0',
      method: 'eth_getLogs',
      params: [params],
      id: `${cfg.address}_${cfg.event}`,
    }
    return result
  }

  async processEvents(events: any[]) {
    let blockNumber = 0
    if (events.length === 0) {
      return blockNumber
    }
    logger.info(`process events: ${events.length}`)
    for (const event of events) {
      const address = events[0].address
      const topic = events[0].topics[0]
      const cfg = this.processer.get((address + topic).toLowerCase())
      if (!cfg) {
        continue
      }
      event.chain = this.chainCfg.id + ''
      event.event = cfg.event
      let result = decodeEvent(cfg.abi, event)
      // cfg.fromBlock = Math.max (parseInt(event.blockNumber, 16) + 1, cfg.fromBlock)
      event.decodedData = result
      let res = await retryEthGetBlockByNumber(this.rpc, event.blockNumber, false)
      event.blockTime = parseInt(res.result.timestamp, 16)
      event.dateTag = formatDate(new Date(event.blockTime * 1000))
      const record = await GeneralEvent.saveEvent(event)
      blockNumber = Math.max(parseInt(event.blockNumber, 16), cfg.fromBlock)
      if (cfg.eventProcesser && eventProcessers[cfg.eventProcesser]) {
        try {
          await eventProcessers[cfg.eventProcesser].parseEvent(record)
        } catch (err) {
          logger.error('process event error: ', err.message || err)
        }
      }
    }
    return blockNumber
  }
}
